import numpy as np

from adam_core.coordinates import CartesianCoordinates, Origin
from adam_core.orbits import Orbits
from adam_core.time import Timestamp

from ..utils import ensure_input_origin_and_frame, ensure_input_time_scale


def test_ensure_input_time_scale():
    # Test that ensure_input_time_scale works
    times = Timestamp.from_mjd(np.array([1.0, 2.0, 3.0]), scale="tdb")
    orbits = Orbits.from_kwargs(
        coordinates=CartesianCoordinates.from_kwargs(
            x=np.array([1.0, 2.0, 3.0]),
            y=np.array([1.0, 2.0, 3.0]),
            z=np.array([1.0, 2.0, 3.0]),
            vx=np.array([1.0, 2.0, 3.0]),
            vy=np.array([1.0, 2.0, 3.0]),
            vz=np.array([1.0, 2.0, 3.0]),
            time=times,
        ),
    )

    results = orbits.set_column(
        "coordinates.time", orbits.coordinates.time.rescale("utc")
    )
    assert results.coordinates.time.scale == "utc"

    results = ensure_input_time_scale(orbits, times)
    assert results.coordinates.time.scale == "tdb"


def test_ensure_input_origin_and_frame():
    # Test that ensure_input_origin_and_frame works
    times = Timestamp.from_mjd(np.array([60000.0, 60001.0, 60002.0]), scale="tdb")
    orbits = Orbits.from_kwargs(
        coordinates=CartesianCoordinates.from_kwargs(
            x=np.array([1.0, 2.0, 3.0]),
            y=np.array([1.0, 2.0, 3.0]),
            z=np.array([1.0, 2.0, 3.0]),
            vx=np.array([1.0, 2.0, 3.0]),
            vy=np.array([1.0, 2.0, 3.0]),
            vz=np.array([1.0, 2.0, 3.0]),
            time=times,
            origin=Origin.from_kwargs(code=["EARTH", "SUN", "MOON"]),
            frame="equatorial",
        ),
    )

    results = Orbits.from_kwargs(
        orbit_id=orbits.orbit_id,
        coordinates=CartesianCoordinates.from_kwargs(
            x=orbits.coordinates.x,
            y=orbits.coordinates.y,
            z=orbits.coordinates.z,
            vx=orbits.coordinates.vx,
            vy=orbits.coordinates.vy,
            vz=orbits.coordinates.vz,
            time=orbits.coordinates.time,
            origin=Origin.from_kwargs(
                code=["SOLAR_SYSTEM_BARYCENTER", "SUN", "SOLAR_SYSTEM_BARYCENTER"]
            ),
            frame="ecliptic",
        ),
    )

    results = ensure_input_origin_and_frame(orbits, results)
    assert results.coordinates.origin.code.to_pylist() == ["EARTH", "SUN", "MOON"]
    assert results.coordinates.frame == "equatorial"



import quivr as qv
import spiceypy as sp

from ...constants import KM_P_AU, S_P_DAY
from ...time import Timestamp
from ...utils.spice import get_perturber_state, setup_SPICE
from ..cartesian import CartesianCoordinates
from ..origin import Origin, OriginCodes
from ..transform import cartesian_to_frame, transform_coordinates
from .test_transforms_translation import assert_coords_equal


def test_cartesian_to_frame(orbital_elements, orbital_elements_equatorial):
    # Test cartesian_to_frame correctly converts between ecliptic and
    # equatorial cartesian coordinates

    cartesian_coordinates_ecliptic = CartesianCoordinates.from_kwargs(
        x=orbital_elements["x"].values,
        y=orbital_elements["y"].values,
        z=orbital_elements["z"].values,
        vx=orbital_elements["vx"].values,
        vy=orbital_elements["vy"].values,
        vz=orbital_elements["vz"].values,
        time=Timestamp.from_mjd(orbital_elements["mjd_tdb"].values, scale="tdb"),
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN" for i in range(len(orbital_elements))]),
    )

    cartesian_coordinates_equatorial = CartesianCoordinates.from_kwargs(
        x=orbital_elements_equatorial["x"].values,
        y=orbital_elements_equatorial["y"].values,
        z=orbital_elements_equatorial["z"].values,
        vx=orbital_elements_equatorial["vx"].values,
        vy=orbital_elements_equatorial["vy"].values,
        vz=orbital_elements_equatorial["vz"].values,
        time=Timestamp.from_mjd(
            orbital_elements_equatorial["mjd_tdb"].values, scale="tdb"
        ),
        frame="equatorial",
        origin=Origin.from_kwargs(
            code=["SUN" for i in range(len(orbital_elements_equatorial))]
        ),
    )

    # Convert equatorial coordinates to ecliptic
    cartesian_coordinates_ecliptic_actual = cartesian_to_frame(
        cartesian_coordinates_equatorial,
        "ecliptic",
    )
    assert cartesian_coordinates_ecliptic_actual.frame == "ecliptic"

    # Test that the two coordinates are equal to 10 mm in position and 10 nm/s in velocity
    assert_coords_equal(
        cartesian_coordinates_ecliptic_actual, cartesian_coordinates_ecliptic
    )

    # Convert ecliptic coordinates to equatorial
    cartesian_coordinates_equatorial_actual = cartesian_to_frame(
        cartesian_coordinates_ecliptic,
        "equatorial",
    )
    assert cartesian_coordinates_equatorial_actual.frame == "equatorial"

    # Test that the two coordinates are equal to 10 mm in position and 10 nm/s in velocity
    assert_coords_equal(
        cartesian_coordinates_equatorial_actual, cartesian_coordinates_equatorial
    )


def test_transform_coordinates_frame(orbital_elements, orbital_elements_equatorial):
    # Test transform_coordinates correctly converts between ecliptic and
    # equatorial cartesian coordinates

    cartesian_coordinates_ecliptic = CartesianCoordinates.from_kwargs(
        x=orbital_elements["x"].values,
        y=orbital_elements["y"].values,
        z=orbital_elements["z"].values,
        vx=orbital_elements["vx"].values,
        vy=orbital_elements["vy"].values,
        vz=orbital_elements["vz"].values,
        time=Timestamp.from_mjd(
            orbital_elements["mjd_tdb"].values,
            scale="tdb",
        ),
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN" for i in range(len(orbital_elements))]),
    )

    cartesian_coordinates_equatorial = CartesianCoordinates.from_kwargs(
        x=orbital_elements_equatorial["x"].values,
        y=orbital_elements_equatorial["y"].values,
        z=orbital_elements_equatorial["z"].values,
        vx=orbital_elements_equatorial["vx"].values,
        vy=orbital_elements_equatorial["vy"].values,
        vz=orbital_elements_equatorial["vz"].values,
        time=Timestamp.from_mjd(
            orbital_elements_equatorial["mjd_tdb"].values,
            scale="tdb",
        ),
        frame="equatorial",
        origin=Origin.from_kwargs(
            code=["SUN" for i in range(len(orbital_elements_equatorial))]
        ),
    )

    # Convert equatorial coordinates to ecliptic
    cartesian_coordinates_ecliptic_actual = transform_coordinates(
        cartesian_coordinates_equatorial,
        frame_out="ecliptic",
    )
    assert cartesian_coordinates_ecliptic_actual.frame == "ecliptic"

    # Test that the two coordinates are equal to 10 mm in position and 10 nm/s in velocity
    assert_coords_equal(
        cartesian_coordinates_ecliptic_actual, cartesian_coordinates_ecliptic
    )

    # Convert ecliptic coordinates to equatorial
    cartesian_coordinates_equatorial_actual = transform_coordinates(
        cartesian_coordinates_ecliptic,
        frame_out="equatorial",
    )
    assert cartesian_coordinates_equatorial_actual.frame == "equatorial"

    # Test that the two coordinates are equal to 10 mm in position and 10 nm/s in velocity
    assert_coords_equal(
        cartesian_coordinates_equatorial_actual, cartesian_coordinates_equatorial
    )


def test_transform_coordinates_to_itrf93():
    """
    Test that transform_coordinates correctly converts between ecliptic and
    ITRF93 cartesian coordinates
    """
    setup_SPICE()

    # Get perturber states for a few of the planets
    times = Timestamp.from_mjd([59000, 59500, 60000], scale="tdb")
    states = CartesianCoordinates.empty()
    target_ids = []
    for perturber in [
        OriginCodes.MOON,
        OriginCodes.VENUS,
        OriginCodes.MARS_BARYCENTER,
        OriginCodes.JUPITER_BARYCENTER,
        OriginCodes.SATURN_BARYCENTER,
        OriginCodes.URANUS_BARYCENTER,
        OriginCodes.NEPTUNE_BARYCENTER,
    ]:
        states = qv.concatenate(
            [
                states,
                get_perturber_state(
                    perturber, times, origin=OriginCodes.SUN, frame="ecliptic"
                ),
            ]
        )
        target_ids.extend(perturber.name for _ in range(len(times)))

    states_itrf93 = transform_coordinates(states, frame_out="itrf93")

    # Repeat with SPICE
    states_spice_itrf93 = CartesianCoordinates.empty()
    for coord, target_id in zip(states_itrf93, target_ids):
        # Rotate the coordinates to the ITRF93 frame using SPICE
        state, lt = sp.spkezr(
            target_id, coord.time.et()[0].as_py(), "ITRF93", "NONE", "SUN"
        )
        states_spice_itrf93 = qv.concatenate(
            [
                states_spice_itrf93,
                CartesianCoordinates.from_kwargs(
                    x=[state[0] / KM_P_AU],
                    y=[state[1] / KM_P_AU],
                    z=[state[2] / KM_P_AU],
                    vx=[state[3] / KM_P_AU * S_P_DAY],
                    vy=[state[4] / KM_P_AU * S_P_DAY],
                    vz=[state[5] / KM_P_AU * S_P_DAY],
                    time=coord.time,  # Add time from original coordinate
                    frame="itrf93",  # Specify the frame
                    origin=Origin.from_kwargs(code=["SUN"]),  # Specify the origin
                ),
            ]
        )

    # Test that the two coordinate sets are equal within tolerance
    assert_coords_equal(
        states_itrf93, states_spice_itrf93, position_tol_mm=10, velocity_tol_nm_s=500
    )


def test_transform_coordinates_from_itrf93():
    """
    Test that transform_coordinates correctly converts between ITRF93 and
    ecliptic cartesian coordinates
    """
    setup_SPICE()

    # Get perturber states for a few of the planets
    times = Timestamp.from_mjd([59000, 59500, 60000], scale="tdb")
    states = CartesianCoordinates.empty()
    target_ids = []
    for perturber in [
        OriginCodes.MOON,
        OriginCodes.VENUS,
        OriginCodes.MARS_BARYCENTER,
        OriginCodes.JUPITER_BARYCENTER,
        OriginCodes.SATURN_BARYCENTER,
        OriginCodes.URANUS_BARYCENTER,
        OriginCodes.NEPTUNE_BARYCENTER,
    ]:
        states = qv.concatenate(
            [
                states,
                get_perturber_state(
                    perturber, times, origin=OriginCodes.SUN, frame="itrf93"
                ),
            ]
        )
        target_ids.extend(perturber.name for _ in range(len(times)))

    states_ecliptic = transform_coordinates(states, frame_out="ecliptic")

    # Repeat with SPICE
    states_spice_ecliptic = CartesianCoordinates.empty()
    for coord, target_id in zip(states_ecliptic, target_ids):
        # Rotate the coordinates to the ecliptic frame using SPICE
        state, lt = sp.spkezr(
            target_id, coord.time.et()[0].as_py(), "ECLIPJ2000", "NONE", "SUN"
        )
        states_spice_ecliptic = qv.concatenate(
            [
                states_spice_ecliptic,
                CartesianCoordinates.from_kwargs(
                    x=[state[0] / KM_P_AU],
                    y=[state[1] / KM_P_AU],
                    z=[state[2] / KM_P_AU],
                    vx=[state[3] / KM_P_AU * S_P_DAY],
                    vy=[state[4] / KM_P_AU * S_P_DAY],
                    vz=[state[5] / KM_P_AU * S_P_DAY],
                    time=coord.time,  # Add time from original coordinate
                    frame="ecliptic",  # Specify the frame
                    origin=Origin.from_kwargs(code=["SUN"]),  # Specify the origin
                ),
            ]
        )

    # Test that the two coordinate sets are equal within tolerance
    assert_coords_equal(
        states_ecliptic,
        states_spice_ecliptic,
        position_tol_mm=10,
        velocity_tol_nm_s=500,
    )



import numpy as np
import pytest
import quivr as qv

from ...coordinates.keplerian import KeplerianCoordinates
from ...coordinates.origin import Origin, OriginCodes
from ...orbits import Orbits
from ...time import Timestamp
from ...utils.spice import get_perturber_state
from ..moid import calculate_moid, calculate_perturber_moids


def test_calculate_moid_circular_orbits():
    # Test for two coplanar circular orbits that their MOID
    # is equal to the difference of their semi-major axes
    circular_1 = Orbits.from_kwargs(
        orbit_id=["a"],
        coordinates=KeplerianCoordinates.from_kwargs(
            a=[1.0],
            e=[0.0],
            i=[0.0],
            ap=[0.0],
            raan=[0.0],
            M=[0.0],
            time=Timestamp.from_mjd([59000.0], scale="tdb"),
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"]),
        ).to_cartesian(),
    )

    circular_3 = Orbits.from_kwargs(
        orbit_id=["b"],
        coordinates=KeplerianCoordinates.from_kwargs(
            a=[3.0],
            e=[0.0],
            i=[0.0],
            ap=[0.0],
            raan=[0.0],
            M=[0.0],
            time=Timestamp.from_mjd([59000.0], scale="tdb"),
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"]),
        ).to_cartesian(),
    )

    moid_13, moid_time_13 = calculate_moid(circular_1, circular_3)
    np.testing.assert_allclose(moid_13, 2.0, rtol=0, atol=1e-12)

    moid_31, moid_time_31 = calculate_moid(circular_3, circular_1)
    np.testing.assert_allclose(moid_31, 2.0, rtol=0, atol=1e-12)

    np.testing.assert_allclose(moid_13, moid_31, rtol=0, atol=1e-12)

    # Test that the MOID of two orbits with relative inclination of 90 degrees
    # is equal to the difference of their semi-major axes
    circular_1 = Orbits.from_kwargs(
        orbit_id=["a"],
        coordinates=KeplerianCoordinates.from_kwargs(
            a=[1.0],
            e=[0.0],
            i=[0.0],
            ap=[0.0],
            raan=[0.0],
            M=[0.0],
            time=Timestamp.from_mjd([59000.0], scale="tdb"),
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"]),
        ).to_cartesian(),
    )

    inclined_3 = Orbits.from_kwargs(
        orbit_id=["b"],
        coordinates=KeplerianCoordinates.from_kwargs(
            a=[3.0],
            e=[0.0],
            i=[90.0],
            ap=[0.0],
            raan=[0.0],
            M=[0.0],
            time=Timestamp.from_mjd([59000.0], scale="tdb"),
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"]),
        ).to_cartesian(),
    )

    moid_13, moid_time_13 = calculate_moid(circular_1, inclined_3)
    np.testing.assert_allclose(moid_13, 2.0, rtol=0, atol=1e-12)

    moid_31, moid_time_31 = calculate_moid(inclined_3, circular_1)
    np.testing.assert_allclose(moid_31, 2.0, rtol=0, atol=1e-12)

    np.testing.assert_allclose(moid_13, moid_31, rtol=0, atol=1e-12)


def test_calculate_moid_noncircular_orbits():

    # Test that the MOID of two coplanar but non-circular orbits located on
    # the x-axis at same time is equal to the difference
    # of their periapse distances
    j2000 = Timestamp.from_jd([2451545.0], scale="tt")

    noncircular_keplerian_1 = KeplerianCoordinates.from_kwargs(
        a=[2.0],
        e=[0.2],
        i=[0.0],
        ap=[0.0],
        raan=[0.0],
        M=[0.0],
        time=j2000,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"]),
    )
    noncircular_1 = Orbits.from_kwargs(
        orbit_id=["a"],
        coordinates=noncircular_keplerian_1.to_cartesian(),
    )

    noncircular_keplerian_2 = KeplerianCoordinates.from_kwargs(
        a=[5.0],
        e=[0.2],
        i=[0.0],
        ap=[0.0],
        raan=[0.0],
        M=[0.0],
        time=j2000,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"]),
    )
    noncircular_2 = Orbits.from_kwargs(
        orbit_id=["b"],
        coordinates=noncircular_keplerian_2.to_cartesian(),
    )

    moid_13, moid_time_13 = calculate_moid(noncircular_1, noncircular_2)
    np.testing.assert_allclose(
        moid_13,
        noncircular_keplerian_2.q[0] - noncircular_keplerian_1.q[0],
        rtol=0,
        atol=1e-12,
    )

    moid_31, moid_time_31 = calculate_moid(noncircular_2, noncircular_1)
    np.testing.assert_allclose(
        moid_31,
        noncircular_keplerian_2.q[0] - noncircular_keplerian_1.q[0],
        rtol=0,
        atol=1e-12,
    )

    # Test again with two orbits that overlap at some point
    noncircular_keplerian_1 = KeplerianCoordinates.from_kwargs(
        a=[2.0],
        e=[0.1],
        i=[0.0],
        ap=[0.0],
        raan=[0.0],
        M=[0.0],
        time=j2000,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"]),
    )
    noncircular_1 = Orbits.from_kwargs(
        orbit_id=["a"],
        coordinates=noncircular_keplerian_1.to_cartesian(),
    )

    noncircular_keplerian_2 = KeplerianCoordinates.from_kwargs(
        a=[3.0],
        e=[0.9],
        i=[0.0],
        ap=[0.0],
        raan=[0.0],
        M=[0.0],
        time=j2000,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"]),
    )
    noncircular_2 = Orbits.from_kwargs(
        orbit_id=["b"],
        coordinates=noncircular_keplerian_2.to_cartesian(),
    )

    moid_13, moid_time_13 = calculate_moid(noncircular_1, noncircular_2)
    np.testing.assert_allclose(moid_13, 0.0, rtol=0, atol=1e-6)

    moid_31, moid_time_31 = calculate_moid(noncircular_2, noncircular_1)
    np.testing.assert_allclose(moid_31, 0.0, rtol=0, atol=1e-6)


@pytest.mark.parametrize("max_processes", [1, 2])
def test_calculate_perturber_moids(max_processes):
    # Test that the MOID we can multiprocess the moid calculation
    pertubers = [OriginCodes.EARTH, OriginCodes.MARS_BARYCENTER, OriginCodes.MERCURY]

    time = Timestamp.from_mjd([59000.0], scale="tdb")

    earth = Orbits.from_kwargs(
        orbit_id=["input_EARTH"],
        coordinates=get_perturber_state(OriginCodes.EARTH, time),
    )
    mars = Orbits.from_kwargs(
        orbit_id=["input_MARS"],
        coordinates=get_perturber_state(OriginCodes.MARS_BARYCENTER, time),
    )
    orbits = qv.concatenate([earth, mars])

    moids = calculate_perturber_moids(
        orbits, pertubers, max_processes=max_processes, chunk_size=1
    )

    assert len(moids) == 6
    assert (
        len(moids.select("orbit_id", "input_EARTH").select("perturber.code", "EARTH"))
        == 1
    )
    assert (
        len(
            moids.select("orbit_id", "input_EARTH").select(
                "perturber.code", "MARS_BARYCENTER"
            )
        )
        == 1
    )
    assert (
        len(moids.select("orbit_id", "input_MARS").select("perturber.code", "EARTH"))
        == 1
    )
    assert (
        len(
            moids.select("orbit_id", "input_MARS").select(
                "perturber.code", "MARS_BARYCENTER"
            )
        )
        == 1
    )
    assert (
        len(moids.select("orbit_id", "input_EARTH").select("perturber.code", "MERCURY"))
        == 1
    )
    assert (
        len(moids.select("orbit_id", "input_MARS").select("perturber.code", "MERCURY"))
        == 1
    )



"""
Photometry regression tests using observations and orbits from the MPC and JPL Horizons databases.

Overview
--------
These tests load frozen "fixtures" (.npz files) containing observed magnitudes, observatory
metadata, and precise heliocentric geometry for specific objects. For each fixture, they:

1. Predict magnitudes using the `magnitude` model.
2. Calculate residuals (predicted - observed) using absolute magnitude (H) values from:
   - MPC (stored in `H_v_mpc`)
   - JPL SBDB (stored in `H_v_jpl`)
3. Validate that current residuals do not exceed the baseline residual ceilings (median absolute
   and 95th percentile absolute) stored within the fixture.

Data Sources
------------
- Orbits: JPL Horizons
- Observations: MPC database (pulled 01/13/2026)

Benchmark Results
-----------------
A larger benchmark of 1000 observations spanning at least 10 objects (where possible) gave the
following average residuals for each filter:

filter_id |    n | med_abs_mpc | p95_abs_mpc | mean_mpc | med_abs_jpl | p95_abs_jpl | mean_jpl
----------+------+-------------+-------------+----------+-------------+-------------+---------
ATLAS_c   | 1000 |      0.1494 |      0.5156 |   0.0155 |      0.1469 |      0.5193 |   0.0162
ATLAS_o   | 2000 |      0.2687 |      0.5828 |   0.2408 |      0.2704 |      0.5822 |   0.2440
DECam_VR  |  490 |      0.2016 |     80.3238 |   8.0554 |      0.1761 |      0.5677 |   0.0107
DECam_Y   |   24 |      0.4604 |      0.7152 |   0.3850 |      0.4614 |      0.7172 |   0.3866
DECam_g   | 1000 |      0.2571 |      0.6291 |  -0.2106 |      0.2562 |      0.6282 |  -0.2108
DECam_i   | 1000 |      0.2706 |      0.6542 |   0.2299 |      0.2659 |      0.6472 |   0.2261
DECam_r   | 1000 |      0.1782 |      0.7725 |   0.1441 |      0.1835 |      0.7730 |   0.1464
DECam_u   |   13 |      1.3321 |      1.4790 |  -1.3035 |      1.3411 |      1.4880 |  -1.3125
DECam_z   | 1000 |      0.3547 |      0.7997 |   0.2579 |      0.3507 |      0.7998 |   0.2519
ZTF_g     | 1000 |      0.2275 |      0.5885 |  -0.2308 |      0.2294 |      0.5905 |  -0.2318
ZTF_i     | 1000 |      0.2520 |      0.5271 |   0.2529 |      0.2487 |      0.5267 |   0.2510
ZTF_r     | 1000 |      0.1928 |      0.5008 |   0.1622 |      0.1896 |      0.5028 |   0.1634

Most residuals are on the order of 0.2-0.3 mag, with the exception of DECam_u and DECam_Y for which
there are very few observations.

Note that some objects in the MPC database have their H values incorrectly set to 99.99, causing unusually
high residuals for those objects.

"""

from __future__ import annotations

from pathlib import Path

import numpy as np
import pyarrow as pa
import pytest

from adam_core.coordinates.cartesian import CartesianCoordinates
from adam_core.coordinates.origin import Origin
from adam_core.observations.exposures import Exposures
from adam_core.photometry.bandpasses import map_to_canonical_filter_bands
from adam_core.photometry.magnitude import predict_magnitudes
from adam_core.photometry.tests.data.fixture_generation import (
    observers_from_heliocentric_positions,
)
from adam_core.time import Timestamp

DATA_DIR = Path(__file__).parent / "data"

# Filled in once fixtures are generated (see `generate_mpc_magnitude_fixtures.py`).
# For multi-station benchmarking, auto-discover fixtures on disk.
FIXTURES: list[str] = sorted(
    p.name for p in DATA_DIR.glob("mpc_magnitude_fixture_*.npz")
)
if not FIXTURES:
    FIXTURES = ["__NO_FIXTURES__"]


def _format_residual_table(rows: list[dict[str, object]], cols: list[str]) -> str:
    # Compute widths
    widths: dict[str, int] = {c: len(c) for c in cols}
    for r in rows:
        for c in cols:
            widths[c] = max(widths[c], len(str(r.get(c, ""))))

    def fmt_row(r: dict[str, object]) -> str:
        parts = [str(r.get(c, "")).rjust(widths[c]) for c in cols]
        return "  ".join(parts)

    header = "  ".join([c.rjust(widths[c]) for c in cols])
    sep = "  ".join(["-" * widths[c] for c in cols])
    body = "\n".join(fmt_row(r) for r in rows)
    return "\n".join([header, sep, body])


def _phase_angle_deg(object_pos: np.ndarray, observer_pos: np.ndarray) -> np.ndarray:
    """
    Phase angle (Sun-object-observer) in degrees, matching the H-G geometry used by the model.

    Inputs are heliocentric vectors in AU with shape (N, 3).
    """
    obj = np.asarray(object_pos, dtype=np.float64)
    obs = np.asarray(observer_pos, dtype=np.float64)
    if obj.ndim != 2 or obs.ndim != 2 or obj.shape != obs.shape or obj.shape[1] != 3:
        raise ValueError("object_pos and observer_pos must both have shape (N, 3)")

    r = np.sqrt(np.sum(obj * obj, axis=1))
    delta_vec = obj - obs
    delta = np.sqrt(np.sum(delta_vec * delta_vec, axis=1))
    obs_sun = np.sqrt(np.sum(obs * obs, axis=1))
    numer = r**2 + delta**2 - obs_sun**2
    denom = 2.0 * r * delta
    cos_phase = np.clip(numer / denom, -1.0, 1.0)
    return np.degrees(np.arccos(cos_phase))


def _slope_abs_resid_vs_phase(resid: np.ndarray, phase_deg: np.ndarray) -> float:
    x = np.asarray(phase_deg, dtype=np.float64)
    y = np.asarray(np.abs(resid), dtype=np.float64)
    m = np.isfinite(x) & np.isfinite(y)
    if int(np.sum(m)) < 2:
        return float("nan")
    x = x[m]
    y = y[m]
    x0 = x - float(np.mean(x))
    denom = float(np.sum(x0 * x0))
    if denom <= 0.0:
        return float("nan")
    return float(np.sum(x0 * (y - float(np.mean(y)))) / denom)


def _slope_resid_vs_phase(resid: np.ndarray, phase_deg: np.ndarray) -> float:
    x = np.asarray(phase_deg, dtype=np.float64)
    y = np.asarray(resid, dtype=np.float64)
    m = np.isfinite(x) & np.isfinite(y)
    if int(np.sum(m)) < 3:
        return float("nan")
    x = x[m]
    y = y[m]
    if float(np.ptp(x)) < 1e-6:
        return float("nan")
    x0 = x - float(np.mean(x))
    denom = float(np.sum(x0 * x0))
    if denom <= 0.0:
        return float("nan")
    return float(np.sum(x0 * (y - float(np.mean(y)))) / denom)


def _pearson_r(resid: np.ndarray, phase_deg: np.ndarray) -> float:
    x = np.asarray(phase_deg, dtype=np.float64)
    y = np.asarray(resid, dtype=np.float64)
    m = np.isfinite(x) & np.isfinite(y)
    if int(np.sum(m)) < 3:
        return float("nan")
    x = x[m]
    y = y[m]
    if float(np.ptp(x)) < 1e-6:
        return float("nan")
    x0 = x - float(np.mean(x))
    y0 = y - float(np.mean(y))
    sx = float(np.sqrt(np.sum(x0 * x0)))
    sy = float(np.sqrt(np.sum(y0 * y0)))
    if sx <= 0.0 or sy <= 0.0:
        return float("nan")
    return float(np.sum(x0 * y0) / (sx * sy))


@pytest.mark.parametrize("fixture_name", FIXTURES)
def test_mpc_magnitude_regression_fixture(
    monkeypatch, fixture_name: str, pytestconfig
) -> None:
    if fixture_name == "__NO_FIXTURES__":
        pytest.skip("No MPC photometry fixtures found on disk.")

    fixture_path = DATA_DIR / fixture_name
    if not fixture_path.exists():
        pytest.skip(
            f"Missing fixture {fixture_name}. Generate it with "
            f"`python -m adam_core.photometry.tests.data.generate_mpc_magnitude_fixtures --overwrite`"
        )

    fx = np.load(fixture_path, allow_pickle=True)
    if "H_v_mpc" not in fx.files or "H_v_jpl" not in fx.files:
        pytest.skip(f"Legacy fixture missing dual-H fields: {fixture_name}")
    h_source = (
        str(pytestconfig.getoption("--photometry-fixtures-h-source")).strip().lower()
    )
    if h_source not in {"both", "mpc"}:
        raise ValueError(f"Invalid --photometry-fixtures-h-source: {h_source}")

    station = str(fx["station"][0])
    H_v_mpc = float(fx["H_v_mpc"][0])
    G_mpc = float(fx["G_mpc"][0])
    H_v_jpl = float(fx["H_v_jpl"][0])
    G_jpl = float(fx["G_jpl"][0])
    if not (np.isfinite(H_v_mpc) and np.isfinite(H_v_jpl)):
        pytest.skip(
            f"Missing H in MPC or JPL for fixture object {fx['object_id'][0]}; skipping."
        )
    time_iso = fx["time_iso"].astype(object).tolist()
    bands = fx["band"].astype(object).tolist()
    mag_obs = np.asarray(fx["mag_obs"], dtype=np.float64)
    obj_pos = np.asarray(fx["object_pos"], dtype=np.float64)
    obs_pos = np.asarray(fx["observer_pos"], dtype=np.float64)

    baseline_keys = fx["baseline_keys"].astype(object).tolist()
    baseline_mpc_median_abs = np.asarray(
        fx["baseline_mpc_median_abs"], dtype=np.float64
    )
    baseline_mpc_p95_abs = np.asarray(fx["baseline_mpc_p95_abs"], dtype=np.float64)
    baseline_jpl_median_abs = np.asarray(
        fx["baseline_jpl_median_abs"], dtype=np.float64
    )
    baseline_jpl_p95_abs = np.asarray(fx["baseline_jpl_p95_abs"], dtype=np.float64)

    n = len(mag_obs)
    assert obj_pos.shape == (n, 3)
    assert obs_pos.shape == (n, 3)
    assert len(bands) == n

    times = Timestamp.from_iso8601(time_iso, scale="utc")

    exposures = Exposures.from_kwargs(
        id=[f"e{i}" for i in range(n)],
        start_time=times,
        duration=np.zeros(n, dtype=np.float64),
        filter=bands,
        observatory_code=[station] * n,
    )

    canonical_filter_ids = map_to_canonical_filter_bands(
        exposures.observatory_code,
        exposures.filter,
        allow_fallback_filters=False,
    )
    exposures = exposures.set_column(
        "filter", pa.array(canonical_filter_ids.tolist(), type=pa.large_string())
    )

    observers = observers_from_heliocentric_positions(
        station_code=station,
        times_utc=times,
        heliocentric_pos_au=obs_pos,
    )
    monkeypatch.setattr(exposures, "observers", lambda *args, **kwargs: observers)

    object_coords = CartesianCoordinates.from_kwargs(
        x=obj_pos[:, 0],
        y=obj_pos[:, 1],
        z=obj_pos[:, 2],
        vx=np.zeros(n),
        vy=np.zeros(n),
        vz=np.zeros(n),
        time=times,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"] * n),
    )

    out_mpc = predict_magnitudes(
        H=H_v_mpc,
        object_coords=object_coords,
        exposures=exposures,
        G=G_mpc,
        reference_filter="V",
        composition="NEO",
    )

    resid_mpc = np.asarray(out_mpc, dtype=np.float64) - mag_obs
    resid_jpl: np.ndarray | None
    if h_source == "both":
        out_jpl = predict_magnitudes(
            H=H_v_jpl,
            object_coords=object_coords,
            exposures=exposures,
            G=G_jpl,
            reference_filter="V",
            composition="NEO",
        )
        resid_jpl = np.asarray(out_jpl, dtype=np.float64) - mag_obs
    else:
        resid_jpl = None
    canon = np.asarray(canonical_filter_ids, dtype=object)
    phase_deg = _phase_angle_deg(obj_pos, obs_pos)

    rows: list[dict[str, object]] = []
    for i, k in enumerate(baseline_keys):
        m0_mpc = float(baseline_mpc_median_abs[i])
        p0_mpc = float(baseline_mpc_p95_abs[i])
        m0_jpl = float(baseline_jpl_median_abs[i])
        p0_jpl = float(baseline_jpl_p95_abs[i])
        xs_mpc = resid_mpc[canon == str(k)]
        xs_jpl = (
            resid_jpl[canon == str(k)]
            if resid_jpl is not None
            else np.asarray([], dtype=np.float64)
        )
        ph = phase_deg[canon == str(k)]
        if xs_mpc.size == 0 and (resid_jpl is None or xs_jpl.size == 0):
            continue

        xs_mpc_f = np.asarray(xs_mpc, dtype=np.float64)
        xs_jpl_f = np.asarray(xs_jpl, dtype=np.float64)
        ph_f = np.asarray(ph, dtype=np.float64)

        med_abs_mpc = (
            float(np.abs(np.median(xs_mpc_f))) if xs_mpc_f.size else float("nan")
        )
        p95_abs_mpc = (
            float(np.quantile(np.abs(xs_mpc_f), 0.95))
            if xs_mpc_f.size
            else float("nan")
        )
        mean_mpc = float(np.mean(xs_mpc_f)) if xs_mpc_f.size else float("nan")

        med_abs_jpl = (
            float(np.abs(np.median(xs_jpl_f))) if xs_jpl_f.size else float("nan")
        )
        p95_abs_jpl = (
            float(np.quantile(np.abs(xs_jpl_f), 0.95))
            if xs_jpl_f.size
            else float("nan")
        )
        mean_jpl = float(np.mean(xs_jpl_f)) if xs_jpl_f.size else float("nan")

        d_med = med_abs_jpl - med_abs_mpc
        d_p95 = p95_abs_jpl - p95_abs_mpc

        row: dict[str, object] = {
            "filter_id": str(k),
            "n": int(max(xs_mpc_f.size, xs_jpl_f.size)),
            "phase_deg_span": f"{float(np.ptp(ph_f)):.2f}" if ph_f.size else "nan",
            "slope_resid_mpc(mag/deg)": (
                f"{_slope_resid_vs_phase(xs_mpc_f, ph_f):+.4f}"
                if xs_mpc_f.size
                else "nan"
            ),
            "pearson_r_mpc(resid,phase)": (
                f"{_pearson_r(xs_mpc_f, ph_f):+.3f}" if xs_mpc_f.size else "nan"
            ),
            "median_abs_mpc": f"{med_abs_mpc:.6f} (base {m0_mpc:.6f})",
            "p95_abs_mpc": f"{p95_abs_mpc:.6f} (base {p0_mpc:.6f})",
            "mean_mpc": f"{mean_mpc:.6f}",
        }
        if resid_jpl is not None:
            row.update(
                {
                    "slope_resid_jpl(mag/deg)": (
                        f"{_slope_resid_vs_phase(xs_jpl_f, ph_f):+.4f}"
                        if xs_jpl_f.size
                        else "nan"
                    ),
                    "pearson_r_jpl(resid,phase)": (
                        f"{_pearson_r(xs_jpl_f, ph_f):+.3f}" if xs_jpl_f.size else "nan"
                    ),
                    "median_abs_jpl": f"{med_abs_jpl:.6f} (base {m0_jpl:.6f})",
                    "p95_abs_jpl": f"{p95_abs_jpl:.6f} (base {p0_jpl:.6f})",
                    "mean_jpl": f"{mean_jpl:.6f}",
                    "delta_median_abs(jpl-mpc)": f"{d_med:+.6f}",
                    "delta_p95_abs(jpl-mpc)": f"{d_p95:+.6f}",
                }
            )
        rows.append(row)

    rows.sort(key=lambda r: str(r["filter_id"]))
    if resid_jpl is None:
        cols = [
            "filter_id",
            "n",
            "phase_deg_span",
            "slope_resid_mpc(mag/deg)",
            "pearson_r_mpc(resid,phase)",
            "median_abs_mpc",
            "p95_abs_mpc",
            "mean_mpc",
        ]
    else:
        cols = [
            "filter_id",
            "n",
            "phase_deg_span",
            "slope_resid_mpc(mag/deg)",
            "pearson_r_mpc(resid,phase)",
            "median_abs_mpc",
            "p95_abs_mpc",
            "mean_mpc",
            "slope_resid_jpl(mag/deg)",
            "pearson_r_jpl(resid,phase)",
            "median_abs_jpl",
            "p95_abs_jpl",
            "mean_jpl",
            "delta_median_abs(jpl-mpc)",
            "delta_p95_abs(jpl-mpc)",
        ]
    table = _format_residual_table(rows, cols)

    # Always print the table so runs can be diffed against previous baselines without
    # failing the test due to residual changes.
    if resid_jpl is None:
        print(
            f"\nResidual summary for {fixture_name} (station={station}, object={fx['object_id'][0]}): "
            f"H_mpc={H_v_mpc:.3f}\n{table}\n"
        )
    else:
        print(
            f"\nResidual summary for {fixture_name} (station={station}, object={fx['object_id'][0]}): "
            f"H_mpc={H_v_mpc:.3f}, H_jpl={H_v_jpl:.3f}\n{table}\n"
        )



import numpy as np
import spiceypy as sp
from adam_assist import ASSISTPropagator

from ...coordinates.cartesian import CartesianCoordinates
from ...coordinates.origin import Origin
from ...time import Timestamp
from ..orbits import Orbits
from ..spice_kernel import orbits_to_spk


def test_orbits_to_spk(tmp_path):
    # Create test orbit
    t0 = Timestamp.from_mjd([60000.0], scale="tdb")
    origin = Origin.from_kwargs(code=["SOLAR_SYSTEM_BARYCENTER"])

    # Create a simple circular orbit
    orbits = Orbits.from_kwargs(
        orbit_id=["test_orbit"],
        object_id=["test_object"],
        coordinates=CartesianCoordinates.from_kwargs(
            x=[1.0],
            y=[1.0],
            z=[0.0],
            vx=[0.0],
            vy=[1.0],
            vz=[0.0],
            time=t0,
            origin=origin,
            frame="equatorial",
        ),
    )

    # Generate SPK file
    spk_file = tmp_path / "test.bsp"
    end_time = t0.add_fractional_days(100.0)

    orbits_to_spk(
        orbits,
        str(spk_file),
        start_time=t0,
        end_time=end_time,
        propagator=ASSISTPropagator(),
        step_days=1.0,
        window_days=10.0,
    )

    # Verify file exists
    assert spk_file.exists()

    # Load SPK and verify contents
    sp.furnsh(str(spk_file))

    # Check state at t0
    et0 = t0.et()[0].as_py()
    state, lt = sp.spkez(1000000, et0, "J2000", "NONE", 0)

    # Convert SPICE state (km, km/s) to our units (au, au/day)
    state[:3] /= 149597870.7  # km to au
    state[3:] *= 86400.0 / 149597870.7  # km/s to au/day

    # Compare with input state
    input_state = orbits.coordinates.values[0]
    assert np.allclose(state, input_state, rtol=1e-10, atol=1e-10)

    # Clean up
    sp.unload(str(spk_file))


# def test_chebyshev_fit_against_spice_earth():

#     de440_reader = DE440Reader(de440)
#     x_coeffs, y_coeffs, z_coeffs = de440_reader.get_emb_chebyshev(0.0)
#     print(f"x_coeffs: {x_coeffs}")
#     print(f"y_coeffs: {y_coeffs}")
#     print(f"z_coeffs: {z_coeffs}")

#     # Get the summary record for the handle. dafgs does not
#     # take handle as an input

#     # === Use your function to fit Chebyshev polynomials ===
#     # For example, get Earth state at points within this record's time window:
#     num_points = 100
#     times = Timestamp.from_et(np.linspace(record_start_et, record_end_et, num_points))
#     earth_coordinates = get_perturber_state(
#         perturber=OriginCodes.EARTH_MOON_BARYCENTER,
#         times=times,
#         origin=OriginCodes.SOLAR_SYSTEM_BARYCENTER,
#         frame="equatorial"
#     )

#     # Fit Chebyshev polynomials using your function.
#     # (Assume that fit_chebyshev returns coefficients, a mid time, and a half-interval.)
#     fitted_coeffs, mid_time, half_interval = fit_chebyshev(
#         earth_coordinates,
#         record_start_et,
#         record_end_et,
#         degree=degree
#     )

#     # Compare only the position coefficients (first 3 components)
#     pos_diff = np.abs(fitted_coeffs[:3] - spice_coeffs[:3])
#     print("Position coefficient differences:", pos_diff)

#     # Assert that the position coefficients match within 1 meter (1e-3 km)
#     assert np.all(pos_diff < 1e-3)



import os

import numpy as np
import pyarrow.compute as pc
import pytest

# Check if required modules are available
try:
    from adam_assist import ASSISTPropagator

    from ...orbits.oem_io import orbit_from_oem, orbit_to_oem, orbit_to_oem_propagated

    _OEM_AVAILABLE = True
except ImportError:
    _OEM_AVAILABLE = False

from ...coordinates import CartesianCoordinates
from ...coordinates.covariances import CoordinateCovariances
from ...coordinates.origin import Origin
from ...time import Timestamp
from ..orbits import Orbits

astroforge_optical_path = (
    f"{os.path.dirname(__file__)}/testdata/AstroForgeSC_Optical.oem"
)

covariance_example_path = f"{os.path.dirname(__file__)}/testdata/CovarianceExample.oem"


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_from_oem():
    optical_orbits = orbit_from_oem(astroforge_optical_path)

    first_segment = optical_orbits.apply_mask(
        pc.match_substring(optical_orbits.orbit_id, "seg_0")
    )
    second_segment = optical_orbits.apply_mask(
        pc.match_substring(optical_orbits.orbit_id, "seg_1")
    )

    assert len(first_segment) == 10
    assert len(second_segment) == 5
    assert len(optical_orbits) == 15

    assert optical_orbits.coordinates.frame == "equatorial"  # J2000
    assert optical_orbits.coordinates.origin.code[0].as_py() == "EARTH"


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_from_oem_covariance():
    orbits = orbit_from_oem(covariance_example_path)

    assert len(orbits) == 4

    assert not orbits.coordinates.covariance[0].is_all_nan()
    assert orbits.coordinates.covariance[1:].is_all_nan()


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_to_oem_propagated(tmp_path):
    mjds = [60000.0, 60001.0, 60002.0, 60003.0, 60004.0]
    times = Timestamp.from_mjd(mjds, scale="tdb")
    test_orbit = Orbits.from_kwargs(
        object_id=["object_id1"],
        orbit_id=["test_orbit"],
        coordinates=CartesianCoordinates.from_kwargs(
            time=times[0],
            x=[1.0],
            y=[1.1],
            z=[1.2],
            vx=[0.01],
            vy=[0.01],
            vz=[0.01],
            frame="equatorial",
            origin=Origin.from_kwargs(code=["SUN"]),
            covariance=CoordinateCovariances.from_sigmas(
                np.array([[1.0, 2.0, 3.0, 4.0, 5.0, 6.0]])
            ),
        ),
    )

    oem_path = f"{tmp_path}/test.oem"
    orbit_to_oem_propagated(test_orbit, oem_path, times, ASSISTPropagator)

    # test load
    orbits_rt = orbit_from_oem(oem_path)

    orbits_rt_df = orbits_rt.to_dataframe()

    assert orbits_rt.coordinates.frame == "equatorial"
    assert orbits_rt.coordinates.origin.code[0].as_py() == "SUN"
    assert len(orbits_rt_df) == 5
    assert orbits_rt.coordinates.time.scale == "tdb"
    assert not np.any([cov.is_all_nan() for cov in orbits_rt.coordinates.covariance])


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_to_oem_basic(tmp_path):
    """Test basic functionality of orbit_to_oem with multiple time points."""
    mjds = [60000.0, 60001.0, 60002.0]
    times = Timestamp.from_mjd(mjds, scale="tdb")

    # Create pre-propagated orbits with multiple time points
    test_orbits = Orbits.from_kwargs(
        object_id=["test_object", "test_object", "test_object"],
        orbit_id=["orbit_1", "orbit_2", "orbit_3"],
        coordinates=CartesianCoordinates.from_kwargs(
            time=times,
            x=[1.0, 1.1, 1.2],
            y=[2.0, 2.1, 2.2],
            z=[3.0, 3.1, 3.2],
            vx=[0.01, 0.011, 0.012],
            vy=[0.02, 0.021, 0.022],
            vz=[0.03, 0.031, 0.032],
            frame="equatorial",
            origin=Origin.from_kwargs(code=["SUN", "SUN", "SUN"]),
            covariance=CoordinateCovariances.nulls(3),
        ),
    )

    oem_path = f"{tmp_path}/test_basic.oem"
    result_path = orbit_to_oem(test_orbits, oem_path)

    assert result_path == oem_path
    assert os.path.exists(oem_path)

    # Verify we can read it back
    orbits_rt = orbit_from_oem(oem_path)
    assert len(orbits_rt) == 3
    assert orbits_rt.coordinates.frame == "equatorial"
    assert orbits_rt.coordinates.origin.code[0].as_py() == "SUN"


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_to_oem_with_covariance(tmp_path):
    """Test orbit_to_oem with covariance data."""
    mjds = [60000.0, 60001.0]
    times = Timestamp.from_mjd(mjds, scale="tdb")

    # Create covariance matrices
    cov_matrices = np.array(
        [
            [
                [1.0, 0.1, 0.2, 0.01, 0.02, 0.03],
                [0.1, 2.0, 0.3, 0.04, 0.05, 0.06],
                [0.2, 0.3, 3.0, 0.07, 0.08, 0.09],
                [0.01, 0.04, 0.07, 0.001, 0.002, 0.003],
                [0.02, 0.05, 0.08, 0.002, 0.002, 0.004],
                [0.03, 0.06, 0.09, 0.003, 0.004, 0.003],
            ],
            [
                [1.1, 0.11, 0.21, 0.011, 0.021, 0.031],
                [0.11, 2.1, 0.31, 0.041, 0.051, 0.061],
                [0.21, 0.31, 3.1, 0.071, 0.081, 0.091],
                [0.011, 0.041, 0.071, 0.0011, 0.0021, 0.0031],
                [0.021, 0.051, 0.081, 0.0021, 0.0021, 0.0041],
                [0.031, 0.061, 0.091, 0.0031, 0.0041, 0.0031],
            ],
        ]
    )

    test_orbits = Orbits.from_kwargs(
        object_id=["test_object", "test_object"],
        orbit_id=["orbit_1", "orbit_2"],
        coordinates=CartesianCoordinates.from_kwargs(
            time=times,
            x=[1.0, 1.1],
            y=[2.0, 2.1],
            z=[3.0, 3.1],
            vx=[0.01, 0.011],
            vy=[0.02, 0.021],
            vz=[0.03, 0.031],
            frame="equatorial",
            origin=Origin.from_kwargs(code=["EARTH", "EARTH"]),
            covariance=CoordinateCovariances.from_matrix(cov_matrices),
        ),
    )

    oem_path = f"{tmp_path}/test_covariance.oem"
    orbit_to_oem(test_orbits, oem_path)

    # Verify covariance is preserved
    orbits_rt = orbit_from_oem(oem_path)
    assert len(orbits_rt) == 2
    assert not orbits_rt.coordinates.covariance[0].is_all_nan()
    assert not orbits_rt.coordinates.covariance[1].is_all_nan()


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_to_oem_custom_originator(tmp_path):
    """Test orbit_to_oem with custom originator."""
    mjds = [60000.0]
    times = Timestamp.from_mjd(mjds, scale="tdb")

    test_orbits = Orbits.from_kwargs(
        object_id=["test_object"],
        orbit_id=["orbit_1"],
        coordinates=CartesianCoordinates.from_kwargs(
            time=times,
            x=[1.0],
            y=[2.0],
            z=[3.0],
            vx=[0.01],
            vy=[0.02],
            vz=[0.03],
            frame="equatorial",
            origin=Origin.from_kwargs(code=["SUN"]),
            covariance=CoordinateCovariances.nulls(1),
        ),
    )

    oem_path = f"{tmp_path}/test_originator.oem"
    custom_originator = "CUSTOM TEST ORIGINATOR"
    orbit_to_oem(test_orbits, oem_path, originator=custom_originator)

    assert os.path.exists(oem_path)

    # Read the file and check originator is set correctly
    # Note: We can't easily verify the originator without parsing the OEM file directly
    # but we can at least verify the file was created successfully
    orbits_rt = orbit_from_oem(oem_path)
    assert len(orbits_rt) == 1


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_to_oem_different_origins(tmp_path):
    """Test orbit_to_oem with different supported origins."""
    mjds = [60000.0, 60001.0]
    times = Timestamp.from_mjd(mjds, scale="tdb")

    # Test different origins
    origins = ["SUN", "EARTH", "MARS", "JUPITER"]

    for origin in origins:
        test_orbits = Orbits.from_kwargs(
            object_id=["test_object", "test_object"],
            orbit_id=["orbit_1", "orbit_2"],
            coordinates=CartesianCoordinates.from_kwargs(
                time=times,
                x=[1.0, 1.1],
                y=[2.0, 2.1],
                z=[3.0, 3.1],
                vx=[0.01, 0.011],
                vy=[0.02, 0.021],
                vz=[0.03, 0.031],
                frame="equatorial",
                origin=Origin.from_kwargs(code=[origin, origin]),
                covariance=CoordinateCovariances.nulls(2),
            ),
        )

        oem_path = f"{tmp_path}/test_{origin.lower()}.oem"
        orbit_to_oem(test_orbits, oem_path)

        # Verify origin is preserved
        orbits_rt = orbit_from_oem(oem_path)
        assert orbits_rt.coordinates.origin.code[0].as_py() == origin


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_to_oem_single_time_warning(tmp_path, caplog):
    """Test that orbit_to_oem issues a warning for single time point."""
    mjds = [60000.0]
    times = Timestamp.from_mjd(mjds, scale="tdb")

    test_orbits = Orbits.from_kwargs(
        object_id=["test_object"],
        orbit_id=["orbit_1"],
        coordinates=CartesianCoordinates.from_kwargs(
            time=times,
            x=[1.0],
            y=[2.0],
            z=[3.0],
            vx=[0.01],
            vy=[0.02],
            vz=[0.03],
            frame="equatorial",
            origin=Origin.from_kwargs(code=["SUN"]),
            covariance=CoordinateCovariances.nulls(1),
        ),
    )

    oem_path = f"{tmp_path}/test_single_time.oem"

    with caplog.at_level("WARNING"):
        orbit_to_oem(test_orbits, oem_path)

    assert "only one time" in caplog.text
    assert "orbit_to_oem_propagated" in caplog.text


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_to_oem_multiple_object_ids_error(tmp_path):
    """Test that orbit_to_oem raises error for multiple object_ids."""
    mjds = [60000.0, 60001.0]
    times = Timestamp.from_mjd(mjds, scale="tdb")

    test_orbits = Orbits.from_kwargs(
        object_id=["object_1", "object_2"],  # Different object IDs
        orbit_id=["orbit_1", "orbit_2"],
        coordinates=CartesianCoordinates.from_kwargs(
            time=times,
            x=[1.0, 1.1],
            y=[2.0, 2.1],
            z=[3.0, 3.1],
            vx=[0.01, 0.011],
            vy=[0.02, 0.021],
            vz=[0.03, 0.031],
            frame="equatorial",
            origin=Origin.from_kwargs(code=["SUN", "SUN"]),
            covariance=CoordinateCovariances.nulls(2),
        ),
    )

    oem_path = f"{tmp_path}/test_error.oem"

    with pytest.raises(AssertionError, match="Only one object_id is supported"):
        orbit_to_oem(test_orbits, oem_path)


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_to_oem_null_object_id_error(tmp_path):
    """Test that orbit_to_oem raises error for null object_id."""
    mjds = [60000.0]
    times = Timestamp.from_mjd(mjds, scale="tdb")

    test_orbits = Orbits.from_kwargs(
        object_id=[None],  # Null object ID
        orbit_id=["orbit_1"],
        coordinates=CartesianCoordinates.from_kwargs(
            time=times,
            x=[1.0],
            y=[2.0],
            z=[3.0],
            vx=[0.01],
            vy=[0.02],
            vz=[0.03],
            frame="equatorial",
            origin=Origin.from_kwargs(code=["SUN"]),
            covariance=CoordinateCovariances.nulls(1),
        ),
    )

    oem_path = f"{tmp_path}/test_null_error.oem"

    with pytest.raises(AssertionError, match="Orbits must specify object_id"):
        orbit_to_oem(test_orbits, oem_path)


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_to_oem_unsupported_origin_error(tmp_path):
    """Test that orbit_to_oem raises error for unsupported origin."""
    mjds = [60000.0]
    times = Timestamp.from_mjd(mjds, scale="tdb")

    test_orbits = Orbits.from_kwargs(
        object_id=["test_object"],
        orbit_id=["orbit_1"],
        coordinates=CartesianCoordinates.from_kwargs(
            time=times,
            x=[1.0],
            y=[2.0],
            z=[3.0],
            vx=[0.01],
            vy=[0.02],
            vz=[0.03],
            frame="equatorial",
            origin=Origin.from_kwargs(code=["UNSUPPORTED_ORIGIN"]),
            covariance=CoordinateCovariances.nulls(1),
        ),
    )

    oem_path = f"{tmp_path}/test_unsupported_origin.oem"

    with pytest.raises(ValueError, match="Unsupported origin code"):
        orbit_to_oem(test_orbits, oem_path)


@pytest.mark.skipif(not _OEM_AVAILABLE, reason="OEM dependencies not installed")
def test_orbit_to_oem_time_sorting(tmp_path):
    """Test that orbit_to_oem properly sorts orbits by time."""
    # Create times out of order
    mjds = [60002.0, 60000.0, 60001.0]
    times = Timestamp.from_mjd(mjds, scale="tdb")

    test_orbits = Orbits.from_kwargs(
        object_id=["test_object", "test_object", "test_object"],
        orbit_id=["orbit_1", "orbit_2", "orbit_3"],
        coordinates=CartesianCoordinates.from_kwargs(
            time=times,
            x=[3.0, 1.0, 2.0],  # Values corresponding to unsorted times
            y=[3.0, 1.0, 2.0],
            z=[3.0, 1.0, 2.0],
            vx=[0.03, 0.01, 0.02],
            vy=[0.03, 0.01, 0.02],
            vz=[0.03, 0.01, 0.02],
            frame="equatorial",
            origin=Origin.from_kwargs(code=["SUN", "SUN", "SUN"]),
            covariance=CoordinateCovariances.nulls(3),
        ),
    )

    oem_path = f"{tmp_path}/test_sorting.oem"
    orbit_to_oem(test_orbits, oem_path)

    # Read back and verify times are sorted
    orbits_rt = orbit_from_oem(oem_path)
    times_rt = orbits_rt.coordinates.time.mjd()

    # Should be sorted: 60000.0, 60001.0, 60002.0
    assert times_rt[0].as_py() == 60000.0
    assert times_rt[1].as_py() == 60001.0
    assert times_rt[2].as_py() == 60002.0

    # Verify corresponding x values are also sorted correctly
    x_values = orbits_rt.coordinates.x
    assert x_values[0].as_py() == pytest.approx(1.0)  # x value for 60000.0
    assert x_values[1].as_py() == pytest.approx(2.0)  # x value for 60001.0
    assert x_values[2].as_py() == pytest.approx(3.0)  # x value for 60002.0
